#version 400

// -------------- Uniforms -----------------
uniform sampler2D texSampler0; // Color Tex
uniform sampler2D texSampler1; // Depth Tex
uniform float width;
uniform float height;
uniform float focusDistance; // Distance to Focus

// heat distortion uniforms
uniform sampler2D rndTexSampler;
uniform sampler2D heatMaskTexSampler;
uniform float time;
uniform float effectScale;
uniform float aspectRatio;

// ---------------- varyings ---------------------
in vec2 varTexCoords;

// ------------- output-color of fragment ------------
out vec4 fragColor;

// ---------------------------------------------------

#define PI    3.14159265

vec2 texel = vec2(1.0/width, 1.0/height);

// Highlights
float lumianceThreshold = 0.5;
float gain = 25.0;

// Sampling Setting
int samples = 3; //samples on the first ring
int rings = 5; //ring count

// camera Settings
float range = 15.0; //focal range
float maxblur = 1.5; //clamp value of max blur

// Bokeh Settings
float bias = 0.7; //edge bias

float namount = 0.00001; //noise amount

bool depthblur = true; //Use weighted neighbouring depth
float blurRadius = 2.0; 




//use weighted neighbouring depth
float bdepth(vec2 coords) 
{
	float depth = 0.0;
	float weights[9];
	vec2 offsets[9]; // pixels around vec2 coords
	
	vec2 offset = vec2(texel.x, texel.y) * blurRadius;
	
	offsets[0] = vec2(-offset.x,-offset.y);
	offsets[1] = vec2( 0.0, -offset.y);
	offsets[2] = vec2( offset.x -offset.y);
	
	offsets[3] = vec2( -offset.x, 0.0);
	offsets[4] = vec2( 0.0,   0.0);
	offsets[5] = vec2( offset.x,  0.0);
	
	offsets[6] = vec2( -offset.x, offset.y);
	offsets[7] = vec2( 0.0,   offset.y);
	offsets[8] = vec2( offset.x,  offset.y);
	
	weights[0] = 1.0/16.0; 	weights[1] = 2.0/16.0;	weights[2] = 1.0/16.0;
	weights[3] = 2.0/16.0;	weights[4] = 4.0/16.0;	weights[5] = 2.0/16.0;
	weights[6] = 1.0/16.0;  weights[7] = 2.0/16.0;	weights[8] = 1.0/16.0;
	
	
	for( int i = 0; i < 9; i++ )
	{
		float tmp = texture2D(texSampler1, coords + offsets[i]).r;
		depth += tmp * weights[i];
	}
	
	return depth;
}

//processing the sample
vec3 color(vec2 coords,float blur) 
{
	vec3 color = vec3(0.0);
	
	color.r = texture2D(texSampler0, coords).r;
	color.g = texture2D(texSampler0, coords).g;
	color.b = texture2D(texSampler0, coords).b;
	
	// Lumiance from color http://blenderartists.org/forum/showthread.php?184102-nicer-and-faster-SSAO/page3
	vec3 lumcoeff = vec3(0.299,0.587,0.114);
	float lum = dot(color.rgb, lumcoeff);
	
	float thresh = max((lum - lumianceThreshold) * gain, 0.0);
	
	return color + mix(vec3(0.0),color, thresh * blur);
}

// Random Noise
// http://stackoverflow.com/questions/4200224/random-noise-functions-for-glsl
vec2 randNoise(in vec2 coord) 
{

	float noiseX = clamp(fract(sin(dot(coord ,vec2(12.9898,78.233))) * 43758.5453),0.0,1.0)*2.0-1.0;
	float noiseY = clamp(fract(sin(dot(coord ,vec2(12.9898,78.233)*2.0)) * 43758.5453),0.0,1.0)*2.0-1.0;
	
	return vec2(noiseX,noiseY);
}

vec2 calcHeatDistortedTexCoords(vec2 coords)
{
	vec4 mask = texture2D(heatMaskTexSampler, coords);
	vec2 offsetCoords = vec2(coords.x * effectScale * aspectRatio, coords.y * effectScale + time);
	vec4 offset = texture2D(rndTexSampler, offsetCoords);
	const float distAmount = 0.01;
	offset.x = (offset.x - 1.0) * distAmount * mask.r;
	offset.y = (offset.y - 1.0) * distAmount * mask.r;
	return coords.xy + offset.xy;
}

void main() 
{
	// 1) heat distortion
	vec2 heatDistTexCoords = calcHeatDistortedTexCoords(varTexCoords);
	
	// 2) depth of field
	float depth = texture2D(texSampler1, heatDistTexCoords).x;
	float blur = 0.0;
	
	// use weighted neighbouring depth
	if (depthblur)
	{
		depth = bdepth(heatDistTexCoords);
	}
	
	// focus at certain distance
	blur = clamp((abs(depth - focusDistance)/range)*100.0, -maxblur, maxblur);
	
	vec2 noise = randNoise(heatDistTexCoords) * namount * blur;
	
	float w = (1.0/width)*blur+noise.x;
	float h = (1.0/height)*blur+noise.y;
	
	vec3 col = texture2D(texSampler0, heatDistTexCoords).rgb;
	float s = 1.0;
	
	int ringsamples;
	
	for (int i = 1; i <= rings; i += 1)
	{   
		ringsamples = i * samples;
		
		for (int j = 0 ; j < ringsamples ; j += 1)
		{      
			float step = PI * 2.0 / float(ringsamples);
			float pw = (cos(float(j) * step) * float(i));
			float ph = (sin(float(j) * step) * float(i));
			col += color(heatDistTexCoords + vec2(pw * w, ph * h), blur) * mix(1.0, (float(i)) / (float(rings)), bias);
			s += 1.0 * mix(1.0, (float(i)) / (float(rings)), bias);
		}
	}
	
	
	col /= s;
	
	// output the final color
	fragColor.rgb = col;
	fragColor.a = 1.0;
}


