#version 400

// ----------------- uniforms and constants ----------------
// material properties
uniform vec3 ambientLightColor;
uniform vec3 emissiveLightColor;
uniform vec4 matDiffuseColor;
uniform vec4 matSpecularColor;
uniform float shininess;

// up to two directional light
const unsigned int DIRLIGHTS = 1;
uniform unsigned int numDirLights;
uniform vec3 dirLightDiffuseColor[DIRLIGHTS];
uniform vec3 dirLightSpecularColor[DIRLIGHTS];

// up to four positional lights; we don't need the position in the fragment shader, attenuation will be passed via varying
const unsigned int POSLIGHTS = 6;
uniform unsigned int numPosLights;
uniform vec3 posLightDiffuseColor[POSLIGHTS];
uniform vec3 posLightSpecularColor[POSLIGHTS];

// up to 2 spotlights
const unsigned int SPOTLIGHTS = 1;
uniform unsigned int numSpotLights;
uniform vec3 spotLightDiffuseColor[SPOTLIGHTS];
uniform vec3 spotLightSpecularColor[SPOTLIGHTS];
uniform float spotLightCosCutoff[SPOTLIGHTS];
uniform float spotLightCosCutoffWidth[SPOTLIGHTS]; ///< width to inner cutoff-radius in cos

uniform sampler2D texSampler0;
uniform sampler2D texSampler1;

// shadow stuff
uniform float pixelSize;
uniform sampler2D shadowSampler0;
uniform bool enableShadowMap0;


uniform int dirLightShadowEnabled;
uniform sampler2D dirLightShadowSampler;

uniform int spotLightShadow0Enabled;
uniform sampler2D spotLightShadow0Sampler;

uniform int spotLightShadow1Enabled;
uniform sampler2D spotLightShadow1Sampler;
 
// PCSS
int BLOCKER_SEARCH_NUM_SAMPLES = 16;
int PCF_NUM_SAMPLES = 16;
float NEAR_PLANE = 3.0;
float LIGHT_WORLD_SIZE = 0.5;
float LIGHT_FRUSTUM_WIDTH = 3.75;
float LIGHT_SIZE_UV = LIGHT_WORLD_SIZE / LIGHT_FRUSTUM_WIDTH; 
float avgBlockerDepth = 0.0;
float numBlockers = 0.0;

// --------------------- varyings ---------------------
in vec2 varTexCoords;
in vec4 varDirLightShadowMapTexCoords;
in vec4 varSpotLightShadowMapTexCoords0;
in vec4 varSpotLightShadowMapTexCoords1;
in vec3 varDirLightL[DIRLIGHTS];			///< to light direction in tangent space
in vec3 varDirLightH[DIRLIGHTS];			///< half vector in tangent space
in vec3 varPosLightL[POSLIGHTS];			///< to light direction in tangent space
in vec3 varPosLightH[POSLIGHTS];			///< half vector in tangent space
in float varPosLightAtt[POSLIGHTS];			///< combined attenuation 
in vec3 varSpotLightL[SPOTLIGHTS];			///< to light direction in tangent space
in vec3 varSpotLightH[SPOTLIGHTS];			///< half vector in tangent space
in float varSpotLightAtt[SPOTLIGHTS];		///< combined attenuation 
in vec3 varSpotLightDir[SPOTLIGHTS];		///< spot light's light-direction in tangent space

// ------------- fragment shader output -----------------
out vec4 fragColor;


// ------------------ functions ----------------


//Parallel plane estimation
float PenumbraSize(float zReceiver, float zBlocker) 
{ 
	return (zReceiver - zBlocker) * LIGHT_WORLD_SIZE / zBlocker;
} 

// We search the shadow map and
// average the depths that are closer to the light source than to
// the point being shaded (the receiver).
void FindBlocker( vec2 uv, float zReceiver ) 
{ 


	/// TODO: const array
	vec2 poissonDisk[] = vec2[]
	(
	 vec2( -0.94201624, -0.39906216 ), 
	 vec2( 0.94558609, -0.76890725 ), 
	 vec2( -0.094184101, -0.92938870 ),
	 vec2( 0.34495938, 0.29387760 ),
	 vec2( -0.91588581, 0.45771432 ), 
	 vec2( -0.81544232, -0.87912464 ), 
	 vec2( -0.38277543, 0.27676845 ),
	 vec2( 0.97484398, 0.75648379 ),
	 vec2( 0.44323325, -0.97511554 ), 
	 vec2( 0.53742981, -0.47373420 ), 
	 vec2( -0.26496911, -0.41893023 ), 
	 vec2( 0.79197514, 0.19090188 ), 
	 vec2( -0.24188840, 0.99706507 ), 
	 vec2( -0.81409955, 0.91437590 ), 
	 vec2( 0.19984126, 0.78641367 ),
	 vec2( 0.14383161, -0.14100790 )
	);

    // This uses similar triangles to compute what  
    // area of the shadow map we should search 
	// The size of the
	// search region depends on the light size and the 
	// receivers distance from the light source.
	
	float searchWidth = (zReceiver - 1.0) * LIGHT_SIZE_UV / zReceiver;
	float blockerSum = 0; 
    numBlockers = 0.0; 
     
	// average the depths that are closer to the light source than to the receiver
    for( int i = 0; i < BLOCKER_SEARCH_NUM_SAMPLES; ++i ) 
    { 

		float shadowMapDepth = texture(shadowSampler0, uv + poissonDisk[i] * searchWidth*2.0).r;
        if ( shadowMapDepth < zReceiver) { 
                blockerSum += shadowMapDepth; 
                numBlockers++; 
            } 
     } 
    avgBlockerDepth = blockerSum / numBlockers; 
}

// PCF Filtering
float PCF_Filter( vec2 uv, float zReceiver, float filterRadiusUV ) 
{ 
	/// TODO: const array
	vec2 poissonDisk[] = vec2[]
	(
	 vec2( -0.94201624, -0.39906216 ), 
	 vec2( 0.94558609, -0.76890725 ), 
	 vec2( -0.094184101, -0.92938870 ),
	 vec2( 0.34495938, 0.29387760 ),
	 vec2( -0.91588581, 0.45771432 ), 
	 vec2( -0.81544232, -0.87912464 ), 
	 vec2( -0.38277543, 0.27676845 ),
	 vec2( 0.97484398, 0.75648379 ),
	 vec2( 0.44323325, -0.97511554 ), 
	 vec2( 0.53742981, -0.47373420 ), 
	 vec2( -0.26496911, -0.41893023 ), 
	 vec2( 0.79197514, 0.19090188 ), 
	 vec2( -0.24188840, 0.99706507 ), 
	 vec2( -0.81409955, 0.91437590 ), 
	 vec2( 0.19984126, 0.78641367 ),
	 vec2( 0.14383161, -0.14100790 )
	);


    float sum = 0.0f; 
	for ( int i = 0; i < PCF_NUM_SAMPLES; ++i ) 
    { 
		
        vec2 offset = poissonDisk[i] * filterRadiusUV; 

		// Bad Filtering
        //sum += (texture(shadowSampler0, uv + offset).r < zReceiver) ? 0.0 : 1.0;
		
		// Better Filtering
		float depthglobal = texture(shadowSampler0, uv + offset).r;
		float lal = (depthglobal < zReceiver) ? 0.0 : 1.0;
		depthglobal = texture(shadowSampler0, uv + offset + vec2(pixelSize,pixelSize)).r;
		lal += (depthglobal < zReceiver) ? 0.0 : 1.0;
		depthglobal = texture(shadowSampler0, uv + offset + vec2(pixelSize,-pixelSize)).r;
		lal += (depthglobal < zReceiver) ? 0.0 : 1.0;
		depthglobal = texture(shadowSampler0, uv + offset + vec2(-pixelSize,pixelSize)).r;
		lal += (depthglobal < zReceiver) ? 0.0 : 1.0;
		depthglobal = texture(shadowSampler0, uv + offset + vec2(-pixelSize,-pixelSize)).r;
		lal += (depthglobal < zReceiver) ? 0.0 : 1.0;
		/*depthglobal = texture(shadowSampler0, uv + offset + vec2(pixelSize,0)).r;
		lal += (depthglobal < zReceiver) ? 0.0 : 1.0;
		depthglobal = texture(shadowSampler0, uv + offset + vec2(0,pixelSize)).r;
		lal += (depthglobal < zReceiver) ? 0.0 : 1.0;
		depthglobal = texture(shadowSampler0, uv + offset + vec2(-pixelSize,0)).r;
		lal += (depthglobal < zReceiver) ? 0.0 : 1.0;
		depthglobal = texture(shadowSampler0, uv + offset + vec2(0,-pixelSize)).r;
		lal += (depthglobal < zReceiver) ? 0.0 : 1.0;*/

		lal = lal / 5.0;
		sum += lal;
    } 
	return sum / PCF_NUM_SAMPLES; 
} 

float PCSS (vec4 tempUv)  
{ 
	vec2 uv = tempUv.xy / tempUv.w; 
	float zReceiver = tempUv.z / tempUv.w; 

	avgBlockerDepth = 0.0;
	numBlockers = 0.0;

	// #1 BLOCKER SEARCH
	FindBlocker(uv, zReceiver ); 
	if( numBlockers < 1 )   
		return 1.0f; //There are no occluders so early out (this saves filtering) 
	

	// #2 PENUMBRA SIZE
	float penumbraRatio = PenumbraSize(zReceiver, avgBlockerDepth);     
	float filterRadiusUV = penumbraRatio * LIGHT_SIZE_UV * NEAR_PLANE / zReceiver; 
     
	// #3 FILTER
	return PCF_Filter( uv, zReceiver, filterRadiusUV ); 
}

float SimpleShadows(vec4 uv, sampler2D sampler)
{
	vec4 shadowWdivide = uv / uv.w ;
	if (shadowWdivide.x < 0.0 || shadowWdivide.x > 1.0 || shadowWdivide.y < 0.0 || shadowWdivide.y > 1.0) // bounds check
	{
		return 0.0f;
	}
	else 
	{
		float distFromLight = texture(sampler, shadowWdivide.xy).r;
		return (distFromLight < shadowWdivide.z) ? 0.0 : 1.0;
	}
}

void main()
{
	float dirShadowFactor = 1.0;
	if(dirLightShadowEnabled == 1)
	{
		dirShadowFactor = SimpleShadows(varDirLightShadowMapTexCoords, dirLightShadowSampler);
	}

	float spotShadowFactor0 = 1.0;
	if(spotLightShadow0Enabled == 1)
	{
		spotShadowFactor0 = SimpleShadows(varSpotLightShadowMapTexCoords0, spotLightShadow0Sampler);
	}

	float spotShadowFactor1 = 1.0;
	if(spotLightShadow1Enabled == 1)
	{
		spotShadowFactor1 = SimpleShadows(varSpotLightShadowMapTexCoords1, spotLightShadow1Sampler);
	}

	vec4 texCol = texture2D(texSampler0, varTexCoords);
	vec3 normal = normalize(texture2D(texSampler1, varTexCoords).xyz  * 2.0 - 1.0);

	// initialize all the colors
	vec3 ambient = ambientLightColor * texCol.xyz;
	vec3 diffuse = vec3(0, 0, 0);
	vec3 specular = vec3(0, 0, 0);

	vec3 toLightDirNrm, halfVector;
	float NdotL, NdotH, specularCoeff, att, angleCos, fade;

	// directional lights contribution
	for (int i=0; i < numDirLights; ++i)
	{
		toLightDirNrm = normalize(varDirLightL[i]);
		halfVector = normalize(varDirLightH[i]);
		NdotL = max(0.0, dot(toLightDirNrm, normal)); // lambertian coefficient
		NdotH = max(0.0, dot(halfVector, normal));
		specularCoeff = pow(NdotH, shininess); // specular coefficient
		
		diffuse += (dirLightDiffuseColor[i] * matDiffuseColor.xyz * NdotL * texCol.xyz) * dirShadowFactor;
		specular += (dirLightSpecularColor[i] * matSpecularColor.xyz * specularCoeff * texCol.xyz) * dirShadowFactor;
	}

	// point lights contribution
	for (int i=0; i < numPosLights; ++i)
	{
		toLightDirNrm = normalize(varPosLightL[i]);
		halfVector = normalize(varPosLightH[i]);
		NdotL = max(0.0, dot(toLightDirNrm, normal)); // lambertian coefficient
		NdotH = max(0.0, dot(halfVector, normal));
		specularCoeff = pow(NdotH, shininess); // specular coefficient
		att = varPosLightAtt[i];

		diffuse += posLightDiffuseColor[i] * matDiffuseColor.xyz * NdotL * texCol.xyz / att;
		specular += posLightSpecularColor[i] * matSpecularColor.xyz * specularCoeff * texCol.xyz / att;
	}

	// spot lights contribution
	for (int i=0; i < numSpotLights; ++i)
	{
		toLightDirNrm = normalize(varSpotLightL[i]);
		vec3 spotLightDirNrm = normalize(varSpotLightDir[i]);
		// test whether or not vertex is in light cone
		angleCos = dot(-spotLightDirNrm, toLightDirNrm);
		if (angleCos > spotLightCosCutoff[i] && dot(normal, toLightDirNrm) > 0)
		{
			halfVector = normalize(varSpotLightH[i]);
			NdotL = max(0.0, dot(toLightDirNrm, normal)); // lambertian coefficient
			NdotH = max(0.0, dot(halfVector, normal));
			specularCoeff = pow(NdotH, shininess); // specular coefficient
			att = varSpotLightAtt[i];
			fade = min(1.0, max(0.0, angleCos - spotLightCosCutoff[i]) / spotLightCosCutoffWidth[i]);

			float minimumShadowFucktor = min(spotShadowFactor0, spotShadowFactor1);
			diffuse += (spotLightDiffuseColor[i] * matDiffuseColor.xyz * NdotL * texCol.xyz * fade / att) * minimumShadowFucktor;
			specular += (spotLightSpecularColor[i] * matSpecularColor.xyz * specularCoeff * texCol.xyz * fade / att) * minimumShadowFucktor;
		}
	}

	// add all together
	//diffuse = diffuse * shadowFactor;
	//specular = specular * shadowFactor;
	fragColor = vec4(ambient + diffuse + specular + emissiveLightColor, 1.0);
}