#version 400 core

// ---------------- Uniforms ----------------
// material properties
uniform vec3 ambientLightColor;
uniform vec3 emissiveLightColor;
uniform vec4 matDiffuseColor;

// up to two directional lights
const unsigned int DIRLIGHTS = 1;
uniform unsigned int numDirLights;
uniform vec3 dirLightDiffuseColor[DIRLIGHTS];

// up to four positional lights
const unsigned int POSLIGHTS = 6;
uniform unsigned int numPosLights;
uniform vec3 posLightPosition[POSLIGHTS];		///< positional-light's position in world-space
uniform vec3 posLightDiffuseColor[POSLIGHTS];
uniform float posLightLinearAttenuation[POSLIGHTS];
uniform float posLightQuadraticAttenuation[POSLIGHTS];

// up to 2 spotlights
const unsigned int SPOTLIGHTS = 1;
uniform unsigned int numSpotLights;
uniform vec3 spotLightPosition[SPOTLIGHTS];		///< spotlight's position in world-space
uniform vec3 spotLightDiffuseColor[SPOTLIGHTS];
uniform float spotLightLinearAttenuation[SPOTLIGHTS];
uniform float spotLightQuadraticAttenuation[SPOTLIGHTS];
uniform vec3 spotLightDirection[SPOTLIGHTS];	///< spotlight's direction in world-space, normalized
uniform float spotLightCosCutoff[SPOTLIGHTS];
uniform float spotLightCosCutoffWidth[SPOTLIGHTS]; ///< width to inner cutoff-radius in cos

uniform sampler2D texSampler0;
uniform sampler2D shadowSampler0;

uniform int dirLightShadowEnabled;
uniform sampler2D dirLightShadowSampler;

uniform int spotLightShadow0Enabled;
uniform sampler2D spotLightShadow0Sampler;

uniform int spotLightShadow1Enabled;
uniform sampler2D spotLightShadow1Sampler;

// ------------ varyings passed from the geometry shader --------------
in vec3 varPositionWS;
in vec2 varTexCoords;
in vec4 varDirLightShadowMapTexCoords;
in vec4 varSpotLightShadowMapTexCoords0;
in vec4 varSpotLightShadowMapTexCoords1;


// ------------ output-color of fragment -------------
out vec4 fragColor;


// ---------------- functions -----------------


float SimpleShadows(vec4 uv, sampler2D sampler)
{
	vec4 shadowWdivide = uv / uv.w ;
	if (shadowWdivide.x < 0.0 || shadowWdivide.x > 1.0 || shadowWdivide.y < 0.0 || shadowWdivide.y > 1.0) // bounds check
	{
		return 0.0f;
	}
	else 
	{
		float distFromLight = texture(sampler, shadowWdivide.xy).r;
		return (distFromLight < shadowWdivide.z) ? 0.0 : 1.0;
	}
}



void main(void) 
{

	float dirShadowFactor = 1.0;
	if(dirLightShadowEnabled == 1)
	{
		dirShadowFactor = SimpleShadows(varDirLightShadowMapTexCoords, dirLightShadowSampler);
	}

	float spotShadowFactor0 = 1.0;
	if(spotLightShadow0Enabled == 1)
	{
		spotShadowFactor0 = SimpleShadows(varSpotLightShadowMapTexCoords0, spotLightShadow0Sampler);
	}

	float spotShadowFactor1 = 1.0;
	if(spotLightShadow1Enabled == 1)
	{
		spotShadowFactor1 = SimpleShadows(varSpotLightShadowMapTexCoords1, spotLightShadow1Sampler);
	}

	vec4 texCol = texture(texSampler0, varTexCoords);
	
	// initialize all the colors
	vec3 ambient = ambientLightColor * texCol.xyz;
	vec3 diffuse = vec3(0, 0, 0);

	vec3 toLight, toLightDirNrm;
	float dist, distSq, att, angleCos, fade;

	// directional lights contribution
	for (int i=0; i < numDirLights; ++i)
	{
		diffuse += (dirLightDiffuseColor[i] * matDiffuseColor.xyz * texCol.xyz) * dirShadowFactor;
	}
	
	// point lights contribution
	for (int i=0; i < numPosLights; ++i)
	{
		toLight = posLightPosition[i] - varPositionWS;
		distSq = dot(toLight, toLight);
		dist = sqrt(distSq);
		
		att = 1.0 + posLightLinearAttenuation[i] * dist + posLightQuadraticAttenuation[i] * distSq;
		diffuse += posLightDiffuseColor[i] * matDiffuseColor.xyz * texCol.xyz / att;
	}
	
	// spot lights contribution
	for (int i=0; i < numSpotLights; ++i)
	{
		toLight = spotLightPosition[i] - varPositionWS;
		toLightDirNrm = normalize(toLight);
		// test whether or not vertex is in light cone
		angleCos = dot(-spotLightDirection[i], toLightDirNrm);
		if (angleCos > spotLightCosCutoff[i])
		{
			distSq = dot(toLight, toLight);
			dist = sqrt(distSq);
			fade = min(1.0, max(0.0, angleCos - spotLightCosCutoff[i]) / spotLightCosCutoffWidth[i]);
			
			float minimumShadowFucktor = min(spotShadowFactor0, spotShadowFactor1);

			att = 1.0 + spotLightLinearAttenuation[i] * dist + spotLightQuadraticAttenuation[i] * distSq;
			diffuse += (spotLightDiffuseColor[i] * matDiffuseColor.xyz * texCol.xyz * fade / att) * minimumShadowFucktor;
		}
	}
	
	// add all together
	//diffuse = diffuse * shadowFactor;
	fragColor = vec4(ambient + diffuse + emissiveLightColor, texCol.a);
}