#version 400 core

// ---------------- Uniforms ----------------
// material properties
uniform vec3 ambientLightColor;
uniform vec3 emissiveLightColor;
uniform vec4 matDiffuseColor;

//// up to two directional lights
//const unsigned int DIRLIGHTS = 1;
//uniform unsigned int numDirLights;
//uniform vec3 dirLightDiffuseColor[DIRLIGHTS];

// up to four positional lights
const unsigned int POSLIGHTS = 3;
uniform unsigned int numPosLights;
uniform vec3 posLightPosition[POSLIGHTS];		///< positional-light's position in world-space
uniform vec3 posLightDiffuseColor[POSLIGHTS];
uniform float posLightLinearAttenuation[POSLIGHTS];
uniform float posLightQuadraticAttenuation[POSLIGHTS];

// up to 2 spotlights
const unsigned int SPOTLIGHTS = 2;
uniform unsigned int numSpotLights;
uniform vec3 spotLightPosition[SPOTLIGHTS];		///< spotlight's position in world-space
uniform vec3 spotLightDiffuseColor[SPOTLIGHTS];
uniform float spotLightLinearAttenuation[SPOTLIGHTS];
uniform float spotLightQuadraticAttenuation[SPOTLIGHTS];
uniform vec3 spotLightDirection[SPOTLIGHTS];	///< spotlight's direction in world-space, normalized
uniform float spotLightCosCutoff[SPOTLIGHTS];
uniform float spotLightCosCutoffWidth[SPOTLIGHTS]; ///< width to inner cutoff-radius in cos

uniform sampler2D texSampler0;

// shadow stuff
uniform vec2 pixelSize;

uniform int dirLightShadowEnabled;
uniform sampler2D dirLightShadowSampler;

uniform int spotLightShadow0Enabled;
uniform sampler2D spotLightShadow0Sampler;

uniform int spotLightShadow1Enabled;
uniform sampler2D spotLightShadow1Sampler;

// PCSS
int BLOCKER_SEARCH_NUM_SAMPLES = 16;
int PCF_NUM_SAMPLES = 16;
float NEAR_PLANE = 3.0;
float LIGHT_WORLD_SIZE = 0.5;
float LIGHT_FRUSTUM_WIDTH = 3.75;
float LIGHT_SIZE_UV = LIGHT_WORLD_SIZE / LIGHT_FRUSTUM_WIDTH; 
float avgBlockerDepth = 0.0;
float numBlockers = 0.0;

// ------------ varyings passed from the geometry shader --------------
in vec3 varPositionWS;
in vec2 varTexCoords;
//in vec4 varDirLightShadowMapTexCoords;
in vec4 varSpotLightShadowMapTexCoords0;
in vec4 varSpotLightShadowMapTexCoords1;
in float alpha;

// ------------ output-color of fragment -------------
out vec4 fragColor;


// ---------------- functions -----------------


//Parallel plane estimation
float PenumbraSize(float zReceiver, float zBlocker) 
{ 
	return (zReceiver - zBlocker) * LIGHT_WORLD_SIZE / zBlocker;
} 

// We search the shadow map and
// average the depths that are closer to the light source than to
// the point being shaded (the receiver).
void FindBlocker( vec2 uv, float zReceiver, sampler2D sampler) 
{ 


	/// TODO: const array
	vec2 poissonDisk[] = vec2[]
	(
	 vec2( -0.94201624, -0.39906216 ), 
	 vec2( 0.94558609, -0.76890725 ), 
	 vec2( -0.094184101, -0.92938870 ),
	 vec2( 0.34495938, 0.29387760 ),
	 vec2( -0.91588581, 0.45771432 ), 
	 vec2( -0.81544232, -0.87912464 ), 
	 vec2( -0.38277543, 0.27676845 ),
	 vec2( 0.97484398, 0.75648379 ),
	 vec2( 0.44323325, -0.97511554 ), 
	 vec2( 0.53742981, -0.47373420 ), 
	 vec2( -0.26496911, -0.41893023 ), 
	 vec2( 0.79197514, 0.19090188 ), 
	 vec2( -0.24188840, 0.99706507 ), 
	 vec2( -0.81409955, 0.91437590 ), 
	 vec2( 0.19984126, 0.78641367 ),
	 vec2( 0.14383161, -0.14100790 )
	);

    // This uses similar triangles to compute what  
    // area of the shadow map we should search 
	// The size of the
	// search region depends on the light size and the 
	// receivers distance from the light source.
	
	float searchWidth = (zReceiver - 1.0) * LIGHT_SIZE_UV / zReceiver; // gute .... why ?
	//float searchWidth = (zReceiver - NEAR_PLANE) * LIGHT_SIZE_UV / zReceiver; // schlechte results
	float blockerSum = 0; 
    numBlockers = 0.0; 
     
	// average the depths that are closer to the light source than to the receiver
    for( int i = 0; i < BLOCKER_SEARCH_NUM_SAMPLES; ++i ) 
    { 

		float shadowMapDepth = texture(sampler, uv + poissonDisk[i] * searchWidth*2.0).r;
        if ( shadowMapDepth < zReceiver) { 
                blockerSum += shadowMapDepth; 
                numBlockers++; 
            } 
     } 
    avgBlockerDepth = blockerSum / numBlockers; 
}

// PCF Filtering
float PCF_Filter( vec2 uv, float zReceiver, float filterRadiusUV, sampler2D sampler) 
{ 
	/// TODO: const array
	vec2 poissonDisk[] = vec2[]
	(
	 vec2( -0.94201624, -0.39906216 ), 
	 vec2( 0.94558609, -0.76890725 ), 
	 vec2( -0.094184101, -0.92938870 ),
	 vec2( 0.34495938, 0.29387760 ),
	 vec2( -0.91588581, 0.45771432 ), 
	 vec2( -0.81544232, -0.87912464 ), 
	 vec2( -0.38277543, 0.27676845 ),
	 vec2( 0.97484398, 0.75648379 ),
	 vec2( 0.44323325, -0.97511554 ), 
	 vec2( 0.53742981, -0.47373420 ), 
	 vec2( -0.26496911, -0.41893023 ), 
	 vec2( 0.79197514, 0.19090188 ), 
	 vec2( -0.24188840, 0.99706507 ), 
	 vec2( -0.81409955, 0.91437590 ), 
	 vec2( 0.19984126, 0.78641367 ),
	 vec2( 0.14383161, -0.14100790 )
	);


    float sum = 0.0f; 
	for ( int i = 0; i < PCF_NUM_SAMPLES; ++i ) 
    { 
		
        vec2 offset = poissonDisk[i] * filterRadiusUV; 

		// Bad Filtering
        //sum += (texture(shadowSampler0, uv + offset).r < zReceiver) ? 0.0 : 1.0;
		
		// Better Filtering
		float depthglobal = texture(sampler, uv + offset).r;
		float lal = (depthglobal < zReceiver) ? 0.0 : 1.0;
		depthglobal = texture(sampler, uv + offset + vec2(pixelSize.x,pixelSize.y)).r;
		lal += (depthglobal < zReceiver) ? 0.0 : 1.0;
		depthglobal = texture(sampler, uv + offset + vec2(pixelSize.x,-pixelSize.y)).r;
		lal += (depthglobal < zReceiver) ? 0.0 : 1.0;
		depthglobal = texture(sampler, uv + offset + vec2(-pixelSize.x,pixelSize.y)).r;
		lal += (depthglobal < zReceiver) ? 0.0 : 1.0;
		depthglobal = texture(sampler, uv + offset + vec2(-pixelSize.x,-pixelSize.y)).r;
		lal += (depthglobal < zReceiver) ? 0.0 : 1.0;
		/*depthglobal = texture(sampler, uv + offset + vec2(pixelSize.x,0)).r;
		lal += (depthglobal < zReceiver) ? 0.0 : 1.0;
		depthglobal = texture(sampler, uv + offset + vec2(0,pixelSize.y)).r;
		lal += (depthglobal < zReceiver) ? 0.0 : 1.0;
		depthglobal = texture(sampler, uv + offset + vec2(-pixelSize.x,0)).r;
		lal += (depthglobal < zReceiver) ? 0.0 : 1.0;
		depthglobal = texture(sampler, uv + offset + vec2(0,-pixelSize.y)).r;
		lal += (depthglobal < zReceiver) ? 0.0 : 1.0;*/

		lal = lal / 5.0;
		sum += lal;
    } 
	return sum / PCF_NUM_SAMPLES; 
} 

float PCSS (vec4 tempUv, sampler2D sampler) 
{ 
	vec2 uv = tempUv.xy / tempUv.w; 
	float zReceiver = tempUv.z / tempUv.w; // Assumed to be eye-space z in this code 
	//float zReceiver = varPositionVS.z / varPositionVS.w; // Assumed to be eye-space z in this code 

	avgBlockerDepth = 0.0;
	numBlockers = 0.0;

	// #1 BLOCKER SEARCH
	FindBlocker(uv, zReceiver, sampler ); 
	if( numBlockers < 1 )   
		return 1.0f; //There are no occluders so early out no need for filtering
	

	// #2 PENUMBRA SIZE
	float penumbraRatio = PenumbraSize(zReceiver, avgBlockerDepth);     
	float filterRadiusUV = penumbraRatio * LIGHT_SIZE_UV * NEAR_PLANE / zReceiver; 
	//float filterRadiusUV = PenumbraSize(zReceiver, avgBlockerDepth);     
     
	// #3 FILTER
	return PCF_Filter( uv, zReceiver, filterRadiusUV, sampler ); 
}



float SimpleShadows(vec4 uv, sampler2D sampler)
{
	vec4 shadowWdivide = uv / uv.w ;
	if (shadowWdivide.x < 0.0 || shadowWdivide.x > 1.0 || shadowWdivide.y < 0.0 || shadowWdivide.y > 1.0) // bounds check
	{
		return 0.0f;
	}
	else 
	{
		float distFromLight = texture(sampler, shadowWdivide.xy).r;
		return (distFromLight < shadowWdivide.z) ? 0.0 : 1.0;
	}
}



void main(void) 
{

	//// TODO: PCSS
	//float dirShadowFactor = 1.0;
	//if(dirLightShadowEnabled == 1)
	//{
	//	dirShadowFactor = SimpleShadows(varDirLightShadowMapTexCoords, dirLightShadowSampler);
	//}

	float spotShadowFactor0 = 1.0;
	if(spotLightShadow0Enabled == 1)
	{
		spotShadowFactor0 = PCSS(varSpotLightShadowMapTexCoords0, spotLightShadow0Sampler);
	}

	float spotShadowFactor1 = 1.0;
	if(spotLightShadow1Enabled == 1)
	{
		spotShadowFactor1 = PCSS(varSpotLightShadowMapTexCoords1, spotLightShadow1Sampler);
	}
	
	vec4 texCol = texture(texSampler0, varTexCoords);
	
	// initialize all the colors
	vec3 ambient = ambientLightColor * texCol.xyz;
	vec3 diffuse = vec3(0, 0, 0);

	vec3 toLight, toLightDirNrm;
	float dist, distSq, att, angleCos, fade;

	//// directional lights contribution
	//for (int i=0; i < numDirLights; ++i)
	//{
	//	diffuse += (dirLightDiffuseColor[i] * matDiffuseColor.xyz * texCol.xyz) * dirShadowFactor;
	//}
	
	// point lights contribution
	for (int i=0; i < numPosLights; ++i)
	{
		toLight = posLightPosition[i] - varPositionWS;
		distSq = dot(toLight, toLight);
		dist = sqrt(distSq);
		
		att = 1.0 + posLightLinearAttenuation[i] * dist + posLightQuadraticAttenuation[i] * distSq;
		diffuse += posLightDiffuseColor[i] * matDiffuseColor.xyz * texCol.xyz / att;
	}
	
	// spot lights contribution
	for (int i=0; i < numSpotLights; ++i)
	{
		toLight = spotLightPosition[i] - varPositionWS;
		toLightDirNrm = normalize(toLight);
		// test whether or not vertex is in light cone
		angleCos = dot(-spotLightDirection[i], toLightDirNrm);
		if (angleCos > spotLightCosCutoff[i])
		{
			distSq = dot(toLight, toLight);
			dist = sqrt(distSq);
			fade = min(1.0, max(0.0, angleCos - spotLightCosCutoff[i]) / spotLightCosCutoffWidth[i]);
			
			att = 1.0 + spotLightLinearAttenuation[i] * dist + spotLightQuadraticAttenuation[i] * distSq;
			
			if (0 == i)
			{
				diffuse += (spotLightDiffuseColor[i] * matDiffuseColor.xyz * texCol.xyz * fade / att) * spotShadowFactor0;
			}
			else
			{
				diffuse += (spotLightDiffuseColor[i] * matDiffuseColor.xyz * texCol.xyz * fade / att) * spotShadowFactor1;
			}
		}
	}
	
	// add all together
	//diffuse = diffuse * shadowFactor;
	fragColor = vec4(ambient + diffuse + emissiveLightColor, texCol.a * alpha);
}