#version 400

// -------------- Uniforms -----------------
// material properties
uniform vec3 ambientLightColor;
//uniform vec3 emissiveLightColor;
//uniform vec4 matDiffuseColor;
uniform vec4 matSpecularColor;
uniform float shininess;
uniform float matReflectFactor; 

// up to two directional light
const unsigned int DIRLIGHTS = 1;
uniform unsigned int numDirLights;
uniform vec3 dirLightDirection[DIRLIGHTS];		///< light-direction in view-space, must be normalized
//uniform vec3 dirLightDiffuseColor[DIRLIGHTS];
uniform vec3 dirLightSpecularColor[DIRLIGHTS];

// up to four positional lights
const unsigned int POSLIGHTS = 4;
uniform unsigned int numPosLights;
uniform vec3 posLightPosition[POSLIGHTS];		///< positional-light's position in view-space
//uniform vec3 posLightDiffuseColor[POSLIGHTS];
uniform vec3 posLightSpecularColor[POSLIGHTS];
uniform float posLightLinearAttenuation[POSLIGHTS];
uniform float posLightQuadraticAttenuation[POSLIGHTS];

// up to 2 spotlights
const unsigned int SPOTLIGHTS = 1;
uniform unsigned int numSpotLights;
uniform vec3 spotLightPosition[SPOTLIGHTS];		///< spotlight's position in view-space
//uniform vec3 spotLightDiffuseColor[SPOTLIGHTS];
uniform vec3 spotLightSpecularColor[SPOTLIGHTS];
uniform float spotLightLinearAttenuation[SPOTLIGHTS];
uniform float spotLightQuadraticAttenuation[SPOTLIGHTS];
uniform vec3 spotLightDirection[SPOTLIGHTS];	///< spotlight's direction in view-space, normalized
uniform float spotLightCosCutoff[SPOTLIGHTS];
uniform float spotLightCosCutoffWidth[SPOTLIGHTS]; ///< width to inner cutoff-radius in cos

uniform sampler2D texSampler0;

// ---------------- varyings ---------------------
in vec4 varVertexPosition;
in vec3 varPositionVS; ///< vertex position in viewing-space
in vec3 varNormalVS;   ///< vertex normal in viewing-space

// ------------- output-color of fragment ------------
out vec4 fragColor;

void main()
{
	vec3 eyePosVS = vec3(0.0, 0.0, 0.0);
	vec3 toEyeDir = normalize(eyePosVS - varPositionVS);
	vec3 normal = normalize(varNormalVS);
	vec2 coords = varVertexPosition.xy / varVertexPosition.w * 0.5 + 0.5;
	vec4 texCol = texture2D(texSampler0, coords);

	// initialize all the colors
	vec3 ambient = ambientLightColor * texCol.xyz;
	vec3 specular = vec3(0, 0, 0);

	vec3 toLight, toLightDirNrm, halfVector;
	float NdotL, NdotH, specularCoeff, dist, distSq, att, angleCos, fade;

	// directional lights contribution
	for (int i=0; i < numDirLights; ++i)
	{
		toLightDirNrm = -dirLightDirection[i];
		halfVector = normalize(toLightDirNrm + toEyeDir);
		NdotL = max(0.0, dot(toLightDirNrm, normal)); // lambertian coefficient
		NdotH = max(0.0, dot(halfVector, normal));
		specularCoeff = pow(NdotH, shininess); // specular coefficient
		
		specular += dirLightSpecularColor[i] * matSpecularColor.xyz * specularCoeff * texCol.xyz;
	}

	// point lights contribution
	for (int i=0; i < numPosLights; ++i)
	{
		toLight = posLightPosition[i] - varPositionVS;
		distSq = dot(toLight, toLight);
		dist = sqrt(distSq);
		toLightDirNrm = normalize(toLight);
		halfVector = normalize(toLightDirNrm + toEyeDir);
		NdotL = max(0.0, dot(toLightDirNrm, normal)); // lambertian coefficient
		NdotH = max(0.0, dot(halfVector, normal));
		specularCoeff = pow(NdotH, shininess); // specular coefficient

		att = 1.0 + posLightLinearAttenuation[i] * dist + posLightQuadraticAttenuation[i] * distSq;
		specular += posLightSpecularColor[i] * matSpecularColor.xyz * specularCoeff * texCol.xyz / att;
	}

	// spot lights contribution
	for (int i=0; i < numSpotLights; ++i)
	{
		toLight = spotLightPosition[i] - varPositionVS;
		toLightDirNrm = normalize(toLight);
		// test whether or not vertex is in light cone
		angleCos = dot(-spotLightDirection[i], toLightDirNrm);
		if (angleCos > spotLightCosCutoff[i] && dot(normal, toLightDirNrm) > 0)
		{
			distSq = dot(toLight, toLight);
			dist = sqrt(distSq);
			halfVector = normalize(toLightDirNrm + toEyeDir);
			NdotL = max(0.0, dot(toLightDirNrm, normal)); // lambertian coefficient
			NdotH = max(0.0, dot(halfVector, normal));
			specularCoeff = pow(NdotH, shininess); // specular coefficient
			fade = min(1.0, max(0.0, angleCos - spotLightCosCutoff[i]) / spotLightCosCutoffWidth[i]);

			att = 1.0 + spotLightLinearAttenuation[i] * dist + spotLightQuadraticAttenuation[i] * distSq;
			specular += spotLightSpecularColor[i] * matSpecularColor.xyz * specularCoeff * texCol.xyz * fade / att;
		}
	}

	// add everything together
	fragColor = vec4(ambient + specular, matReflectFactor);
}